--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local addonName, NS = ...

NS.DebugPrint = FIZZLEDEBUG and FIZZLEDEBUG(AddonName) or function() end

NS.FizzleLib = NS.FizzleLib or {}
local Lib = NS.FizzleLib

local Version = 1
local Compatibility = 1 -- with Version
local path
if addonName == "FizzleLib" then
	path = "Interface\\Addons\\"..addonName.."\\"
else
	path = "Interface\\Addons\\FauxMazzleConfig\\Libs\\FizzleLib\\"
end
Lib.TexturePath = path.."Textures\\"

local band, bor, bxor = bit.band, bit.bor, bit.bxor
local floor = math.floor
local len, sub, find, upper = string.len, string.sub, string.find, string.upper

Lib.Font = "Fonts\\ARIALN.TTF"
Lib.FontSize = 10

Lib.Points = {
	TOP="BOTTOM",
	BOTTOM="TOP",
	LEFT="RIGHT",
	RIGHT="LEFT",
	TOPLEFT="TOPRIGHT",
	TOPRIGHT="TOPLEFT",
	BOTTOMLEFT="BOTTOMRIGHT",
	BOTTOMRIGHT="BOTTOMLEFT",
	CENTER = "CENTER",
}

--------------------------- Misc. -------------------------------------------------
function Lib:GetTexturePath()
	return Lib.TexturePath
end

function Lib.SplitFont(font, size, flags)
	local f, s, fl
	if flags then
		fl = flags
	end
	if size then 
		if type(size) == "number" then
			s = size
		else
			fl = size
		end
	end
	if font then
		if type(font) == "number" then
			s = font
		else
			if Lib.CheckForPath(font) then
				f = font
			else
				fl = font
			end
		end
	end
	return f, s, fl
end

function Lib.BuildFont(font, size, flags)
	f, s, fl = Lib.SplitFont(font, size, flags)
	f = f or Lib.Font
	s = s or Lib.FontSize
	fl = fl or Lib.FontFlags
	return f, s, fl
end

function Lib.GetFont(font, size, flags)
	return  Lib.Font, Lib.FontSize, Lib.FontFlags
end

function Lib.InitData(self, data, index, initfunc)
	if  not index then
		print("FizzleLib InitData!", "No index supplied for:", self:GetName())
		return
	end
	self.Data = data or self.Data or self
	self.Index = index or self.Index
	self.InitFunc = initfunc or self.InitFunc
	if self.InitFunc and not self.init then
		self.Init = function(self, value)
			if value ~= nil then
				self.Data[self.Index] = value
			end
			self:InitFunc()
			if self.UpdateOnInit and self.PostUpdate then
				self:PostUpdate()
			end
		end
		self:Init()
	end
end

---------------------- Edge Border ---------------------------------------------------------
local function ColorBorders(self, r, g, b, a)
	for i=1, 4 do
		self.Borders[i]:SetColorTexture(r, g, b, a and a or 1)
	end		
end

function Lib:CreateBorder(frame, size)
	local f = frame
	local t = size or 1
	f.SetBorderColor = f.SetBorderColor or ColorBorders
	f.Borders = f.Borders or {}
	local offset = (t/2)
	for i=1, 4 do
		if not f.Borders[i] then
			f.Borders[i] = f:CreateLine(nil, "BORDER", nil, 0)
		end
		local b = f.Borders[i]
		b:SetThickness(t)
		b:SetColorTexture(1, 1, 0, 1)
		if i==1 then
			b:SetStartPoint("TOPLEFT", -offset, 0)
			b:SetEndPoint("TOPRIGHT", offset, 0)
		elseif i==2 then
			b:SetStartPoint("TOPRIGHT", 0, offset)
			b:SetEndPoint("BOTTOMRIGHT", 0, -offset)
		elseif i==3 then
			b:SetStartPoint("BOTTOMRIGHT", offset, 0)
			b:SetEndPoint("BOTTOMLEFT", -offset, 0)
		else
			b:SetStartPoint("BOTTOMLEFT", 0, -offset)
			b:SetEndPoint("TOPLEFT", 0, offset)
		end
	end
end

---------------------- Backdrops ---------------------------------------------------------
function Lib:CreateBackdrop(name, parent, ftype, bg, edge, totile, esize, tsize, l, r, t, b, close)
	local prefix = ""
	if parent ~= UIParent then
		prefix = "$parent"
	end
	local f
	if type(name) == "table" then -- add backdrop to existing frame
		f = name
		if not f.SetBackdrop then
			Mixin(f, BackdropTemplateMixin)
		end
	else
		f = CreateFrame(ftype, prefix..name, parent, BackdropTemplateMixin and "BackdropTemplate")
	end
	f:SetSize(5, 5)
	local backdrop = {
		bgFile = bg,
		edgeFile = edge,
		tile = totile,
		edgeSize = esize,
		tileSize = tsize,
		insets = {
			left = l,
			right = r,
			top = t,
			bottom = b,
		},
	}
	f:SetBackdrop(backdrop)
	f:SetBackdropColor(0, 0, 0, 0)
	if close and not f.Close then
		f.Close = Lib:CreateCloseButton("Close", f, nil, nil, close)
	end
	return f
end

function Lib:CreatePlainBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "", true, 1, 0, 0, 0, 0, 0, close)
	return f
end

function Lib:CreateFizzleBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", Lib.TexturePath.."FizzleBorder", true, 1, 8, 1, 1, 1, 1, close)
	if close then
		f.Close:ClearAllPoints()
		f.Close:SetPoint("TOPRIGHT")
	end
	return f
end

function Lib:CreateTooltipBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\Tooltips\\UI-Tooltip-Border", true, 7, 7, 2, 2, 2, 2, close)
	return f
end

function Lib:CreateGlueTooltipBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\GLUES\\Common\\Glue-Tooltip-Border", true, 8, 8, 12, 8, 8, 12, close)
	return f
end

function Lib:CreateLFGBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\LFGFRAME\\LFGBorder", true, 7, 7, 8, 8, 8, 8, close)
	return f
end

function Lib:CreateTextBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\GLUES\\Common\\TextPanel-Border", true, 7, 7, 8, 8, 8, 8, close)
	return f
end

function Lib:CreateAchievmentBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\ACHIEVEMENTFRAME\\UI-Achievement-WoodBorder", true, 7, 7, 8, 8, 8, 8, close)
	return f
end

function Lib:CreateSliderBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\BUTTONS\\UI-SliderBar-Border", true, 7, 7, 8, 8, 8, 8, close)
	return f
end

function Lib:CreateToastBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\FriendsFrame\\UI-Toast-Border", true, 7, 7, 8, 8, 8, 8, close)
	return f
end

function Lib:CreateDropBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", nil, true, 7, 7, 0, 0, 0, 0, close)
	return f
end


function Lib:CreateDialogueBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\DialogFrame\\UI-DialogBox-Border", true, 7, 16, 2, 2, 2, 2, close)
	return f
end

function Lib:CreateGoldDialogueBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\BUTTONS\\WHITE8X8", "Interface\\DialogFrame\\UI-DialogBox-Gold-Border", true, 7, 16, 2, 2, 2, 2, close)
	return f
end

function Lib:CreateDarkBackdrop(name, parent, ftype, close)
	local f = Lib:CreateBackdrop(name, parent, ftype, "Interface\\DialogFrame\\UI-DialogBox-Background-Dark", "", true, 32, 32, 0, 0, 0, 0, close)
	return f
end

function Lib:CreateOptionsHeader(name, parent, ftype, close)
	local f = Lib:CreateDarkBackdrop(name, parent, ftype, close)
	f:SetSize(50, 25)
	f.Close:ClearAllPoints()
	f.Close:SetPoint("TOPRIGHT", 2, 2)
	f.Close:SetPoint("BOTTOMLEFT", f, "BOTTOMRIGHT", -27, -2)
	f:SetBackdropColor(0.6, 0.6, 0.6)
	f.Title = Lib:CreateFont("Title", f, "", "CENTER", "CENTER")
--	f.Title:SetAllPoints()
	f.Title:SetPoint("CENTER")
	function f:SetTitleFont(font, size)
		local Font, Size = self.Title:GetFont()
		Font = font or Size 
		Size = size or Size 
--		if type(font) == "number" then
--			Size = font
--		else
--			Font = font
--		end
		self.Title:SetFont(Font, Size, "")
	end
	function f:SetTitle(text) self.Title:SetText(text) end
	function f:SetTitleColor(r, g, b, a) self.Title:SetTextColor(r, g, b, a) end
	return f
end

function Lib:CreateDragHeader(name, parent, ftype, close)
	local f = Lib:CreateOptionsHeader(name, parent, ftype, close)
	f:EnableMouse(true)
	f:RegisterForDrag("LeftButton")
	f:SetMovable(true) -- Create frame before PLAYER_LOGIN to save position
	f:SetClampedToScreen(true)
	f:SetScript("OnDragStart", function(self) self:StartMoving() end)
	f:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() end)
	return f
end

function Lib:CreateDragBody(name, parent, height)
	local f = Lib:CreateBackdrop(name, parent, "Frame", "Interface\\DialogFrame\\UI-DialogBox-Background", "", true, 32, 32, 0, 0, 0, 0, close)
	f:SetBackdropColor(0.6, 0.6, 0.6)
	if parent and parent ~= UIParent and parent ~= WorldFrame then
		f:SetPoint("TOPLEFT", parent, "BOTTOMLEFT")
		f:SetPoint("TOPRIGHT", parent, "BOTTOMRIGHT")
	end
	f:SetHeight(height or 50)
	return f
end

function Lib:CreateDragFrame(name, parent, width, height, close)
	local f = Lib:CreateDragHeader(name, parent, "Frame", close)
	f:SetWidth(width)
	f.Body = Lib:CreateDragBody("DragBody", f, height)
	return f
end

function Lib:CreateDragSizeFrame(name, parent, width, height, close)
	local n
	if not parent or parent == UIParent then
		n = name
	else
		n = parent:GetName()..name
	end
	local f = Lib:CreateBackdrop(n, parent, "Frame", "Interface\\DialogFrame\\UI-DialogBox-Background", "", true, 32, 32, 0, 0, 0, 0)
	f:SetBackdropColor(0.6, 0.6, 0.6)
	f:SetSize(width, height + 20)
	f:SetMovable(true)
	f:SetResizable(true)
	f:RegisterForDrag("LeftButton")
	f.Dragable = true
	f.Sizeable = true
	function f.SetResizeable(sizable)
		f.Sizeable = sizable
		f.Resize:SetShown(sizable)
	end
	f.Drag = Lib:CreateDarkBackdrop("Drag", f, "Frame", close)
	f.Drag:SetBackdropColor(0, 0, 0)
	f.Drag:SetPoint("TOPLEFT")
	f.Drag:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT", 0, -20)
	f.Drag:EnableMouse(true)
	f.Drag:RegisterForDrag("LeftButton")
	if f.Drag.Close then
		f.Close = f.Drag.Close
		f.Drag.Close:SetScript("OnClick", function(self)
			self:GetParent():GetParent():Hide()
		end)
	end
	f.Drag:SetScript("OnDragStart", function(self)
		local parent = self:GetParent()
		if parent.Dragable then
			parent:StartMoving()
		end
	end)
	f.Drag:SetScript("OnDragStop", function(self)
		local parent = self:GetParent()
		if parent.Dragable then
			parent:StopMovingOrSizing()
		end
	end)

	f.Resize = CreateFrame("Frame", nil, f)
	f.Resize:SetSize(10, 10)
	f.Resize.Texture = f.Resize:CreateTexture()
	f.Resize.Texture:SetAllPoints()
	f.Resize.Texture:SetTexture(Lib.TexturePath.."Resize")
	f.Resize.Texture:SetAlpha(0.7)
	f.Resize:SetPoint("BOTTOMRIGHT")
	f.Resize:EnableMouse(true)
	f.Resize:RegisterForDrag("LeftButton")
	f.Resize:SetScript("OnDragStart", function(self)
		local parent = self:GetParent()
		if parent.Sizeable then
			parent:StartSizing("BOTTOMRIGHT")
		end
	end)
	f.Resize:SetScript("OnDragStop", function(self)
		local parent = self:GetParent()
		if parent.Sizeable then
			parent:StopMovingOrSizing()
		end
	end)
	function f:SetTitle(text) 
		if not f.Drag.Title then
			self.Drag.Title = Lib:CreateFont("Title", self.Drag, "", "CENTER", "CENTER")
			self.Drag.Title:SetAllPoints()
			function self:SetTitleColor(r, g, b, a) 
				self.Drag.Title:SetTextColor(r, g, b, a) 
			end
		end
		self.Drag.Title:SetText(text) 
	end
	return f
end

function Lib.CheckForPath(test)
	if  type(test) == "table" or find(test, "/") or find(test, "\\") then
		return true
	end
end

function Lib:CreateFont(name, parent, text, justifyH, justifyV, font, fontsize, fontflags)
	font = font or Lib.Font
	fontsize = fontsize or Lib.FontSize
	local f = parent:CreateFontString("$parent"..name, "OVERLAY", nil, 4)
	local jh, jv
	if justifyV then
		if type(justifyV) == "number" then
			size = justifyV
		else
			if Lib.CheckForPath(justifyV) then
				fnt = justifyV
			else
				jv = justifyV
			end
		end
	end

	if justifyH then
		if type(justifyH) == "number" then
			size = justifyH
		else
			if Lib.CheckForPath(justifyH) then
				fnt = justifyH
			else
				jh = justifyH
			end
		end
	end
	local fnt, size, flags = Lib.SplitFont(font, fontsize, fontflags)
	jh = jh or "LEFT"
	jv = jv or "CENTER"
	fnt = fnt or Lib.Font
	size = size or Lib.FontSize
	flags = flags or Lib.FontFlags

	if type(fnt) == "string" then
		f:SetFont(fnt, size, flags)
	else
		f:SetFontObject(fnt)
	end
	f:SetJustifyH(jh)
	f:SetJustifyV(jv)
	f:SetTextColor(1, 1, 1, 1)
	f:SetText(text or " ")
	f:GetWidth() -- forces a re-size -- or try placing the fontstring on-screen before setting the font. I think this can fix things not sizing properly right after creation
	return f
end

function Lib:SetEnabled(enable)
	if enable then
		self:SetAlpha(1)
	else
		self:SetAlpha(.5)
	end
	if self.LabelButton then
		self.LabelButton:SetEnabled(enable)
	end
end

function Lib:SetX(point, value)
	local p = upper(point)
	local v = value
	if find(p, "LEFT") then
		v = 0-value
	end
	return v
end

function Lib:SetRelativePoint(self, point, xos, yos)
	point = upper(point)
	xos = xos or 0
	yos = yos or 0
	self:ClearAllPoints()
	self:SetPoint(Lib.Points[point], self:GetParent(), point, xos, yos)
end

function Lib:SetLabel(self, label, point, xos, yos)
	if self.LabelButton and type(self.LabelButton) ~= "table" then
		self.LabelButton = CreateFrame("Button", nil, self)
		self.LabelButton:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		self.LabelButton:GetHighlightTexture():SetBlendMode("ADD")
		function self.GetLabelButton()
			return self.LabelButton
		end
	end
	if not self.Label then
		self.Label = Lib:CreateFont("Label", self)
		function self:GetLabel()
			return self.Label
		end
		function self:SetLabel(text)
			self.Label:SetText(text)
		end
	end
	self.Label:SetText(label)
	if not point then return end
	point = upper(point)
	local opposite = Lib.Points[point]
	self.Label:SetJustifyH(opposite)
	xos = xos or 5
	yos = yos or 0
	xos = Lib:SetX(point, xos)
	Lib:SetRelativePoint(self.Label, point, xos, yos)
	xos = 2
	xos = Lib:SetX(point, xos)
	if self.LabelButton then
		self.LabelButton:ClearAllPoints()
		self.LabelButton:SetPoint("CENTER", self.Label, xos,0)
		self.LabelButton:SetSize(self.Label:GetWidth() + 6, self.Label:GetHeight() + 4)
	end
	self:HookScript("OnEnable", function(self)
		self.Label:SetAlpha(1)
	end)
	self:HookScript("OnDisable", function(self)
		self.Label:SetAlpha(0.5)
	end)
end

function Lib:SetFont(self, font, size, flags)
	if not self.Label then return end
	local ft, sz, fl = Lib.BuildFont(font, size, flags)
	self.Label:SetFont(ft, sz, fl)		
end

function Lib:SetToolTip(self, tip, anchor)
	self.fizzletiptext = tip
	if self.fizzletipinit then return end -- no multiple hooks
	self.fizzletipinit = true
	self:HookScript("OnEnter", function(self)
		if self.fizzletiptext then
			GameTooltip:SetOwner(self, anchor or "ANCHOR_LEFT")
--			GameTooltip:SetText(self.fizzletiptext)
			GameTooltip:SetText(self.fizzletiptext, nil, nil, nil, nil, true)
		end
	end)
	self:HookScript("OnLeave", function(self)
		if self.fizzletiptext then
			GameTooltip:Hide()
		end
	end)
end

function Lib:AddLabelFuncs(self)
	function self:SetLabel(label, point, xos, yos)
		Lib:SetLabel(self, label, point, xos, yos)
	end
	function self:SetFont(font, size, flags)
		Lib:SetFont(self, font, size, flags)
	end
	function self:SetToolTip(tip, anchor)
		Lib:SetToolTip(self, tip, anchor)
	end
end

function Lib:Band(value, flag)
	if bit.band(value, flag) > 0 then
		return true
	else
		return false
	end
end

--[[
local SaveTables = 0x00000001 -- Keep current tables in tact
local SaveKeys = 0x00000010 -- Keep current keys in tact
local DeleteKeys = 0x00000100 -- Remove old keys ie. they don't exist in the source table
]]--
local SaveTables = 0x1 -- Keep current tables in tact
local SaveKeys = 0x2 -- Keep current keys in tact
local DeleteKeys = 0x4 -- Remove old keys ie. they don't exist in the source table


Lib.SaveTables = SaveTables
Lib.SaveKeys = SaveKeys
Lib.DeleteKeys = DeleteKeys
function Lib:CopyTable(source, target, flags) -- preserve retains the integritiy of the table structure
	if flags then
--[[
		local tablepreserve = band(flags, SaveTables)
		local keypreserve = band(flags, SaveKeys)
		local keydelete = band(flags, DeleteKeys)
]]--
		local tablepreserve = band(flags, SaveTables) == SaveTables
		local keypreserve = band(flags, SaveKeys) == SaveKeys
		local keydelete = band(flags, DeleteKeys) == DeleteKeys
	end
	for key, value in pairs(source) do
		if type(value) == "table" then
			if not target[key] or not tablepreserve then
				target[key] = {}
			end
			Lib:CopyTable(value, target[key], flags)
		else
			if not keypreserve or (keypreserve and not target[key]) then
				target[key] = value
			end
		end
	end
	if keydelete then -- delete keys that don't exist in source, no need to check tables that aren't being perserved
		for key, value in pairs(target) do
			if not source[key] then
				target[key] = nil
			end
		end
	end
end

------------------------------------ NEW WORKING VERSION ------------------------------
--Red/Yellow/green basic transition
local Colour_Table = {
	{ pct=0, colour={ r=255, g=0, b=0 } },
	{ pct=50, colour={ r=255, g=225, b=0 } },
	{ pct=100, colour={ r=0, g=255, b=0 } },
}

function Lib:GetColorMix(val, tbl, rangekey, colourkey, divisor)
	local ColourTable = tbl or Colour_Table
	local Key = rangekey or "pct"
	local Colour = colourkey or "colour"
	local Divisor = divisor or 255
	local Found
	for i=2, #ColourTable do
--		Found = i
		if val <= ColourTable[i][Key] then
			Found = i
			break
		end
	end
	if not Found then
		print("FizzleLib:", "Lib:GetColorMix\n", "Invalid value for range:", val)
		return
	end
	local LowKey = ColourTable[Found - 1]
	local HighKey = ColourTable[Found]
	local Range = HighKey[Key] - LowKey[Key]
	local RangePct = (val - LowKey[Key]) / Range
	local PctLower = (1 - RangePct)
	local PctUpper = RangePct
	local r=floor(LowKey[Colour].r * PctLower + HighKey[Colour].r * PctUpper)/Divisor
	local g=floor(LowKey[Colour].g * PctLower + HighKey[Colour].g * PctUpper)/Divisor
	local b=floor(LowKey[Colour].b * PctLower + HighKey[Colour].b * PctUpper)/Divisor
	return r, g, b
end
